
#include "inc.h"

#include "lib.h"
#include "main.h"
#include "div_arm.h"

#include "iwram_arm.h"

//---------------------------------------------------------------------------
//lib.c
extern ST_TIMER Timer;

//main.c
extern ST_BULLET     Bullet[BULLET_MAX_CNT];
extern ST_BULLET_CTL BulletCtl;
//---------------------------------------------------------------------------
ST_DOBULER     Doubler[2];			//_uobt@(\pAp)
ST_DOBULER_CTL DoublerCtl;

u8             NextVCnt;
u32            DummyU32;
//---------------------------------------------------------------------------
IWRAM_CODE void IRQUserHandler()
{
	REG_IME  = IRQ_MASTER_OFF;
	u16 flag = REG_IF;

	if(flag & IRQ_BIT_TIMER0)
	{
		Timer.clockTick += TIMER_INTMS;
	}
	else if(flag & IRQ_BIT_VCOUNT)
	{
		u16 vcount = REG_VCOUNT;

		if(vcount == SCREEN_CY)
		{
			SpriteClear();
			DoublerCtlClear();
			NextVCnt = 228 - DOUBLER_LINE_CY;
		}
		else if(vcount == (228 - DOUBLER_LINE_CY))
		{
			DoublerCopy();
			NextVCnt = 0;
		}
		else
		{
			DoublerCopy();
			NextVCnt += DOUBLER_LINE_CY;
		}
		REG_DISPSTAT = (REG_DISPSTAT & 0x003f) | DSTAT_VCOUNT_LINE(NextVCnt);
	}

	REG_IF  = flag;
	REG_IME = IRQ_MASTER_ON;
}
//---------------------------------------------------------------------------
IWRAM_CODE void SpriteClear()
{
	//obj
	REG_DM0SAD = (u32)&DummyU32;
	REG_DM0DAD = (u32)OAMmem + 8*8;
	REG_DM0CNT = (120 * 4)
	             | ( (DMA_SIZE_16 | DMA_TRANSFER_ON | DMA_TIMING_NOW | DMA_REPEAT_OFF |
	                  DMA_SAD_FIX | DMA_DAD_INC     | DMA_INTR_OFF) << 16 );
}
//---------------------------------------------------------------------------
IWRAM_CODE void DoublerCtlClear()
{
	ST_DOBULER_CTL* p = &DoublerCtl;

	p->copySrc = (u32)&Doubler[p->sw].obj[0];
	p->copyDst = (u32)OAMmem + 8*8;
	p->copyCnt = 0;
	p->vCnt    = 0;
}
//---------------------------------------------------------------------------
IWRAM_CODE void DoublerChange()
{
	DoublerClear(DoublerCtl.sw);
	DoublerCtl.sw ^= 0x01;
}
//---------------------------------------------------------------------------
IWRAM_CODE void DoublerClear(u8 sw)
{
	//obj
	REG_DM1SAD = (u32)&DummyU32;
	REG_DM1DAD = (u32)&Doubler[sw].obj[0];
	REG_DM1CNT = (DOUBLER_BULLET_CNT * 4)
	             | ( (DMA_SIZE_16 | DMA_TRANSFER_ON | DMA_TIMING_NOW | DMA_REPEAT_OFF |
	                  DMA_SAD_FIX | DMA_DAD_INC     | DMA_INTR_OFF) << 16 );
	//objCnt
	REG_DM2SAD = (u32)&DummyU32;
	REG_DM2DAD = (u32)&Doubler[sw].objCnt[0];
	REG_DM2CNT = (DOUBLER_BLOCK_CNT)
	             | ( (DMA_SIZE_16 | DMA_TRANSFER_ON | DMA_TIMING_NOW |  DMA_REPEAT_OFF |
	                  DMA_SAD_FIX | DMA_DAD_INC     | DMA_INTR_OFF) << 16 );
}
//---------------------------------------------------------------------------
//Doubler to OAM
IWRAM_CODE void DoublerCopy()
{
	ST_DOBULER_CTL* p = &DoublerCtl;

	u32 regCnt1 = Doubler[p->sw].objCnt[p->vCnt];
	u32 regCnt2 = 0;

	//DMA]pɃJEg𕪊܂
	if((p->copyCnt + regCnt1) > DOUBLER_OBJ_CNT)
	{
		regCnt2 = (p->copyCnt + regCnt1) - DOUBLER_OBJ_CNT;
		regCnt1 = DOUBLER_OBJ_CNT - p->copyCnt;
	}

	if(regCnt1 > 0)
	{
		REG_DM0SAD = p->copySrc;
		REG_DM0DAD = p->copyDst;
		REG_DM0CNT = (regCnt1 * 4)
		             | ( (DMA_SIZE_16 | DMA_TRANSFER_ON | DMA_TIMING_NOW | DMA_REPEAT_OFF |
		                  DMA_SAD_INC | DMA_DAD_INC     | DMA_INTR_OFF) << 16 );

		p->copySrc += (regCnt1 * 8);
		p->copyDst += (regCnt1 * 8);
		p->copyCnt +=  regCnt1;
	}

	if(regCnt2 > 0)
	{
		REG_DM1SAD = p->copySrc;
		REG_DM1DAD = (u32)OAMmem + 8*8;
		REG_DM1CNT = (regCnt2 * 4)
		             | ( (DMA_SIZE_16 | DMA_TRANSFER_ON | DMA_TIMING_NOW | DMA_REPEAT_OFF |
		                  DMA_SAD_INC | DMA_DAD_INC     | DMA_INTR_OFF) << 16 );

		p->copySrc += (regCnt2 * 8);
		p->copyDst  = (u32)OAMmem + 8*8 + (regCnt2 * 8);
		p->copyCnt  = regCnt2;
	}

	p->vCnt++;
}
//---------------------------------------------------------------------------
IWRAM_CODE void BulletCalc()
{
	u8 sw = DoublerCtl.sw ^ 0x01;

	ST_OBJ* pObj[DOUBLER_BLOCK_CNT];
	s16     i;

	//ꂼ̃ubN̒ɓo^ĂXvCg̐g
	//o^\̃|C^ݒ肵܂
	pObj[0] = &Doubler[sw].obj[0];
	for(i=1; i<DOUBLER_BLOCK_CNT; i++)
	{
		pObj[i] = pObj[i-1] + Doubler[sw].objCnt[i-1];
	}

	//ꂼ̃ubN̒ɃXvCgo^Ă܂
	for(i=0; i<BulletCtl.cnt; i++)
	{
		u8 x      = CommonToFieldX(Bullet[i].x) - 4;	//4́ASHOT_CX / 2
		u8 y      = CommonToFieldY(Bullet[i].y) - 4;
		u8 idx    = y / DOUBLER_LINE_CY;
		ST_OBJ* p = pObj[idx];

		p->attr0 &= 0x1fff;
		p->attr1 &= 0x3fff;
		p->attr0 |= SP_COLOR_16 | SP_SQUARE;
		p->attr1 |= SP_SIZE_8;

		p->attr2 &= 0xfc00;
		p->attr2 |= 70;

		p->attr1 &= 0xfe00;
		p->attr0 &= 0xff00;
		p->attr1 |= (x & 0x01ff);
		p->attr0 |= (y & 0x00ff);

		pObj[idx]++;
	}

	cprint("Bullet = ");
	cprintf(BulletCtl.cnt);
}
//---------------------------------------------------------------------------
IWRAM_CODE void BulletMove()
{
	u8 sw = DoublerCtl.sw ^ 0x01;
	s16 i;

	for(i=0; i<BulletCtl.cnt; i++)
	{
		if(Bullet[i].wait == 0)
		{
			//֐_u|C^gĂ܂iOOG
			(*Bullet[i].pStep)(&Bullet[i]);
		}
		else
		{
			Bullet[i].wait--;

			if(Bullet[i].wait == 0)
			{
				Bullet[i].pStep++;
				i--;
				continue;
			}
		}

		Bullet[i].x += Bullet[i].vx;
		Bullet[i].y += Bullet[i].vy;

		u8 y   = CommonToFieldY(Bullet[i].y) - 4;
		u8 idx = y / DOUBLER_LINE_CY;
		u8 cnt = Doubler[sw].objCnt[idx];

		if( (CommonOut(Bullet[i].x, Bullet[i].y, 4) == TRUE) || (cnt >= DOUBLER_OBJ_CNT) || (y >= 160) )
		{
			Bullet[i].x     = Bullet[BulletCtl.cnt-1].x;
			Bullet[i].y     = Bullet[BulletCtl.cnt-1].y;
			Bullet[i].vx    = Bullet[BulletCtl.cnt-1].vx;
			Bullet[i].vy    = Bullet[BulletCtl.cnt-1].vy;
			Bullet[i].angle = Bullet[BulletCtl.cnt-1].angle;
			Bullet[i].speed = Bullet[BulletCtl.cnt-1].speed;
			Bullet[i].wait  = Bullet[BulletCtl.cnt-1].wait;
			Bullet[i].pStep = Bullet[BulletCtl.cnt-1].pStep;

			BulletCtl.cnt--;
			i--;
			continue;
		}

		Doubler[sw].objCnt[idx]++;
	}
}
//---------------------------------------------------------------------------
inline IWRAM_CODE u16 CommonToFieldX(s16 x)
{
	return (120 + FIX2INT(x));
}
//---------------------------------------------------------------------------
inline IWRAM_CODE u16 CommonToFieldY(s16 y)
{
	return (80 + FIX2INT(y));
}
